package ca.uwaterloo.cs.jgrok.interp;

import java.util.regex.PatternSyntaxException;

public class ValueMath {
    
    public static Value eval(int op, Value left, Value right)
        throws EvaluationException {
        
        switch(op) {
        case Operator.GT:
            return eval_GT(left, right);
            
        case Operator.LT:
            return eval_LT(left, right);
            
        case Operator.EQ:
            return eval_EQ(left, right);
            
        case Operator.NE:
            return eval_NE(left, right);
            
        case Operator.GE:
            return eval_GE(left, right);
            
        case Operator.LE:
            return eval_LE(left, right);
            
        case Operator.ME:
            return eval_ME(left, right);
            
        case Operator.UE:
            return eval_UE(left, right);

        default:
            return Value.VOID;
        }
    }
    
    public static Value eval_GT(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] > floats[1])
                return new Value(true);
            else
                return new Value(false);
        }

        if(sLeft.compareTo(sRight) > 0)
            return new Value(true);
        else
            return new Value(false);
    }
    
    public static Value eval_LT(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] < floats[1])
                return new Value(true);
            else 
                return new Value(false);
        }
        
        if(sLeft.compareTo(sRight) < 0) return new Value(true);
        else return new Value(false);
    }

    public static Value eval_EQ(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] == floats[1])
                return new Value(true);
            else
                return new Value(false);
        }
        
        if(sLeft.compareTo(sRight) == 0)
            return new Value(true);
        else
            return new Value(false);
    }
    
    public static Value eval_NE(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] != floats[1])
                return new Value(true);
            else
                return new Value(false);
        }
        
        if(sLeft.compareTo(sRight) != 0)
            return new Value(true);
        else
            return new Value(false);
    }
    
    public static Value eval_GE(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] >= floats[1])
                return new Value(true);
            else
                return new Value(false);
        }
        
        if(sLeft.compareTo(sRight) >= 0)
            return new Value(true);
        else
            return new Value(false);
    }

    public static Value eval_LE(Value left, Value right) {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        float[] floats = bothFloat(sLeft, sRight);
        if(floats != null) {
            if(floats[0] <= floats[1])
                return new Value(true);
            else
                return new Value(false);
        }
        
        if(sLeft.compareTo(sRight) <= 0)
            return new Value(true);
        else
            return new Value(false);
    }
    
    public static Value eval_ME(Value left, Value right)
        throws EvaluationException {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        try {
            if(sLeft.matches(sRight))
                return new Value(true);
        } catch(PatternSyntaxException e) {
            throw e;
        }
        
        return new Value(false);
    }
    
    public static Value eval_UE(Value left, Value right)
        throws PatternSyntaxException {
        String sLeft, sRight;
        sLeft = left.toString();
        sRight = right.toString();
        
        try {
            if(! sLeft.matches(sRight))
                return new Value(true);
        } catch(PatternSyntaxException e) {
            throw e;
        }
        
        return new Value(false);
    }
    
    @SuppressWarnings("unused")
    private static int[] bothInt(String s1, String s2) {
        try {
            int[] i = new int[2];
            i[0] = Integer.parseInt(s1);
            i[1] = Integer.parseInt(s2);
            return i;
        } catch(NumberFormatException e) {
            return null;
        }
    }
    
    private static float[] bothFloat(String s1, String s2) {
        try {
            float[] f = new float[2];
            f[0] = Float.parseFloat(s1);
            f[1] = Float.parseFloat(s2);
            return f;
        } catch(NumberFormatException e) {
            return null;
        }
    }
}
