package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;

/**
 * <pre>
 * Functions:
 *
 *     NodeSet basket(string items);     // return a singleton.
 *     NodeSet basket(NodeSet itemSet);  // return a singleton.
 *     EdgeSet basket(EdgeSet edgeSet);  // return EdgeSet whose outDegree equals 1.
 *     
 *     NodeSet unbasket(string items);     // reverse basket
 *     NodeSet unbasket(NodeSet itemSet);  // reverse basket
 *     EdgeSet unbasket(EdgeSet edgeSet);  // reverse basket
 *
 * Examples:
 *     >> b1 = basket({1, 2, "ABC"});
 *     >> b2 = basket("(1 2 ABC)");
 *     >> b1
 *     (1 2 ABC)
 *     >> b2
 *     (1 2 ABC)
 *     >> b1 ^ b2
 *     (1 2 ABC)
 *     >> b1 + b2
 *     (1 2 ABC)
 *     >> s = unbasket(b1);
 *     >> s
 *     1
 *     2
 *     ABC
 *     >> r = s X s;
 *     >> basket(r);
 *     1 (1 2 ABC)
 *     2 (1 2 ABC)
 *     ABC (1 2 ABC)
 *     >> unbasket(basket(r));
 *     1 1
 *     1 2
 *     1 ABC
 *     2 1
 *     2 2
 *     2 ABC
 *     ABC 1
 *     ABC 2
 *     ABC ABC
 *     >> 
 * </pre>
 */
public class Basket extends Function {
    
    public Basket() {
        name = "basket";
        type = NodeSet.class;
        paramTypes = new Class[1];
        paramTypes[0] = NodeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        Object o = vals[0].objectValue();
        
        if(getName().equals("basket")) {
            if(o instanceof String) {
                return new Value(SpecialOperation.basket((String)o));
            } else if(o instanceof NodeSet) {
                return new Value(SpecialOperation.basket((NodeSet)o));
            } else {
                return new Value(SpecialOperation.basket((EdgeSet)o));
            }            
        } else if(getName().equals("unbasket")) {
            if(o instanceof String) {
                return new Value(SpecialOperation.unbasket((String)o));
            } else if(o instanceof NodeSet) {
                return new Value(SpecialOperation.unbasket((NodeSet)o));
            } else {
                return new Value(SpecialOperation.unbasket((EdgeSet)o));
            }    
        } else {
            throw InvocationException.createUnknownFunction(getName());
        }
    }
}
