package ca.uwaterloo.cs.jgrok.lib;

import java.io.*;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.interp.Type;
import ca.uwaterloo.cs.jgrok.interp.Value;

public abstract class Function {
    protected String name;
    protected Class<?> type;
    protected Class<?>[] paramTypes;
    private String signature = null;
    
    protected Function() {
        name = null;
        type = void.class;
        paramTypes = new Class[0];
    }
    
    public String getName() {
        return name;
    }
    
    public void setName(String name) {
        this.name = name;
        this.signature = null;
    }
    
    public Class<?> getType() {
        return type;
    }
    
    public void setType(Class<?> type) {
        this.type = type;
        this.signature = null;
    }
    
    public Class<?>[] getParamTypes() {
        return paramTypes;
    }
    
    public void setParamTypes(Class<?>[] paramTypes) {
        this.paramTypes = paramTypes;
        this.signature = null;
    }
    
    public String getSignature() {
        if(signature == null) {
            StringBuffer buffer;
            buffer = new StringBuffer();
            buffer.append(getName());
            
            if(paramTypes == null) {
                buffer.append("(...)");
            } else {
                buffer.append("(");
                for(int i = 0; i < paramTypes.length; i++) {
                    buffer.append(Type.findName(paramTypes[i]));
                    buffer.append(", ");
                }
                
                if(paramTypes.length > 0)
                    buffer.delete(buffer.length()-2, buffer.length());
                
                buffer.append(")");
            }
            
            signature = buffer.toString();
        }
        return signature;
    }
    
    public static String getSignature(String fName, Class<?>[] params) {
        StringBuffer buffer;
        buffer = new StringBuffer(fName);
        
        if(params == null) {
            buffer.append("(...)");
        } else {
            buffer.append("(");
            for(int i = 0; i < params.length; i++) {
                buffer.append(Type.findName(params[i]));
                buffer.append(", ");
            }
            
            if(params.length > 0)
                buffer.delete(buffer.length()-2, buffer.length());
            
            buffer.append(")");
        }
        
        return buffer.toString();
    }
    
    public String synopsis() {
        return Type.findName(getType()) + " " + getSignature();
    }
    
    /**
     * Gets function usage.
     */
    public String usage() {
        String help;
        String infoPath;
        
        help = null;
        infoPath = System.getProperty("ql.info");
        
        if(infoPath != null) {
            try {
                File file;
                String line;
                StringBuffer buffer;
                BufferedReader reader;

                buffer = new StringBuffer(); 
                file = new File(infoPath, name);
                reader = new BufferedReader(new FileReader(file));
                while((line = reader.readLine()) != null) {
                    buffer.append("\n");
                    buffer.append(line);
                }
                
                reader.close();                
                help = buffer.toString();
            } catch(FileNotFoundException e) {
                help = null;
            } catch(IOException e) {
                help = null;
            }
        }
        
        return help;
    }
    
    public String toString() {
        return this.synopsis();
    }
    
    public abstract Value invoke(Env env, Value[] vals)
        throws InvocationException;
}
