package ca.uwaterloo.cs.jgrok.lib;

import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.env.Env;

/**
 * EdgeSet paths(EdgeSet edgeSet)
 * EdgeSet paths(string node, EdgeSet edgeSet)
 * EdgeSet paths(string sourceNode, string targetNode, EdgeSet edgeSet)
 */
public class Paths extends Function {
    
    public Paths() {
        name = "paths";
        type = EdgeSet.class;
        paramTypes = new Class[1];
        paramTypes[0] = EdgeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals) throws InvocationException {
        EdgeSet eSet;
        TupleSet pSet;
        PathClosure pClosure;
        Path[] paths = new Path[0];
        
        if(vals.length == 1) {
            eSet = (EdgeSet)vals[0].objectValue();
            
            pClosure = new PathClosure(eSet);
            paths = pClosure.getPaths();
        } else if(vals.length == 3) {
            String source = vals[0].toString();
            String target = vals[1].toString();
            eSet = (EdgeSet)vals[2].objectValue();
            
            eSet = UtilityOperation.reach(NodeSet.singleton(source),
                                          NodeSet.singleton(target),
                                          eSet);
            
            pClosure = new PathClosure(eSet);
            paths = pClosure.getPaths(source, target);
        }
        
        pSet = new TupleSet(paths.length);
        for(int i = 0; i < paths.length; i++) {
            pSet.add(paths[i].getTuple());
        }
        
        return new Value(pSet);
    }
}
