package ca.uwaterloo.cs.jgrok.lib;

import java.util.*;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;

/**
 * Function <b>sort</b>
 * 
 * Sort set in ascending order.
 * <pre>
 *   TupleSet sort(TupleSet set) 
 *   TupleSet sort(TupleSet set, int colIndex)
 * </pre>
 * 
 * Sort set in descending order.
 * <pre>
 *   TupleSet sortd(TupleSet set)
 *   TupleSet sortd(TupleSet set, int colIndex)
 * </pre>
 * 
 * @author JingweiWu
 */
public class Sort extends Function {
         
    public Sort() {
        name = "sort";
        type = NodeSet.class;
        paramTypes = new Class[1];
        paramTypes[0] = NodeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        Tuple[] data;
        TupleSet tSet;
        TupleSet result;
        boolean ascending = true;
        
        if(name.equals("sort")) ascending = true;
        else if(name.equals("sortd")) ascending = false;
        else ascending = true;
        
        tSet = (TupleSet)vals[0].objectValue();
        tSet = (TupleSet)tSet.clone();
        tSet.removeDuplicates();
        
        result = tSet.newSet();
        data = tSet.getAllTuples();
        
        if(tSet.size() > 0) { 
            if(vals.length == 1) {
                sort(data, result, 0, ascending);
            } else {
                int col = vals[1].intValue();
                if(data[0].size() > col) {
                    sort(data, result, col, ascending);
                } else {
                    throw new InvocationException("index " + col + " out of bounds " + data[0].size());
                }
            }
        }
        
        result.setHasDuplicates(false);
        return new Value(result);
    }
    
    private void sort(Tuple[] data, TupleSet result, int col, boolean upward) {        
        try {
            Arrays.sort(data, new TupleNumCmp(col));
        } catch(NumberFormatException e) {
            Arrays.sort(data, new TupleLexCmp(col));
        }
        
        int count = data.length;
        if(upward)
            for(int i = 0; i < count; i++) {
                result.add(data[i]);
            }
        else
            for(int i = 1; i <= count; i++) {
                result.add(data[count-i]);
            }
    }
    
    class TupleNumCmp implements Comparator<Tuple> {
        int col;
        
        TupleNumCmp(int col) {
            this.col = col;
        }
        
        public int compare(Tuple t1, Tuple t2) throws NumberFormatException {
            float f1 = Float.parseFloat(IDManager.get(t1.get(col)));
            float f2 = Float.parseFloat(IDManager.get(t2.get(col)));
            
            if(f1 < f2) return -1;
            else if(f1 == f2) return 0;
            else return 1;
        }
        
        public boolean equals(Object o) {
            return false;
        }
    }
    
    class TupleLexCmp implements Comparator<Tuple> {
        int col;
        
        TupleLexCmp(int col) {
            this.col = col;
        }
        
        public int compare(Tuple t1, Tuple t2) {
            String s1 = IDManager.get(t1.get(col));
            String s2 = IDManager.get(t2.get(col));
            return s1.compareTo(s2);
        }
        
        public boolean equals(Object o) {
            return false;
        }
    }
}
