package ca.uwaterloo.cs.jgrok.lib;

import java.io.*;
import java.util.Enumeration;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.io.*;

/**
 * Write/append facts to file.
 *
 * <pre>
 * Functions:
 *
 *    void putdb(dataFileName)
 *    void putdb(dataFileName, setToSave)
 *
 *    void putta(dataFileName)
 *    void putta(dataFileName, setToSave)
 *
 *    void appendta(dataFileName, setToSave)
 *    void appenddb(dataFileName, setToSave)
 *
 * </pre>
 */
public class Write extends Function {
    private Env env;
    
    public Write() {
        name = "write";
        type = void.class;
        paramTypes = new Class[1];
        paramTypes[0] = String.class;
    }
    
    public Value invoke(Env env, Value[] vals) throws InvocationException {
        FactbaseWriter writer;
        
        this.env = env;
        if(name.equals("putta")) {
            writer = new TAFileWriter();
            return put(vals, writer);
        } else if(name.equals("putdb")) {
            writer = new RSFFileWriter();
            return put(vals, writer);
        } else if(name.equals("appendta")) {
            return appendta(vals);
        } else if(name.equals("appenddb")) {
            return appenddb(vals);
        } else {
            throw InvocationException.createUnknownFunction(name);
        }
    }
    
    private Value appendta(Value[] vals) throws InvocationException {
        String fileName;
        NodeSet set;
        Node[] nodes;
        
        fileName = vals[0].toString();
        set = (NodeSet)vals[1].objectValue();
        nodes = set.getAllNodes();
        
        Variable var;
        for(int i = 0; i < nodes.length; i++) {
            try {
                var = env.lookup(nodes[i].get());
                if(var.getValue().objectValue() instanceof TupleSet) {
                    TupleSet tSet;
                    tSet = (TupleSet)var.getValue().objectValue();
                    tSet.appendTA(fileName);
                }
            } catch(LookupException e) {
                continue;
            } catch(IOException e2) {
                throw new InvocationException(e2.getMessage());
            }
        }
           
        return Value.VOID;
    }
    
    private Value appenddb(Value[] vals) throws InvocationException {
        String fileName;
        NodeSet set;
        Node[] nodes;
        
        fileName = vals[0].toString();
        set = (NodeSet)vals[1].objectValue();
        nodes = set.getAllNodes();
        
        Variable var;
        for(int i = 0; i < nodes.length; i++) {
            try {
                var = env.lookup(nodes[i].get());
                if(var.getValue().objectValue() instanceof TupleSet) {
                    TupleSet tSet;
                    tSet = (TupleSet)var.getValue().objectValue();
                    tSet.appendDB(fileName);
                }
            } catch(LookupException e) {
                continue;
            } catch(IOException e2) {
                throw new InvocationException(e2.getMessage());
            }
        }
           
        return Value.VOID;
    }
    
    private Value put(Value[] vals, FactbaseWriter writer)
        throws InvocationException {
        Variable var;
        Class<?> type;
        
        String fileName;
        fileName = vals[0].toString();
        Factbase fb = new Factbase();
        if(vals.length == 1) {
            ScriptUnitNode unit;
            Enumeration<Variable> enm;
            
            unit = env.traceScriptUnit();
            enm = unit.allVariables();
            
            while(enm.hasMoreElements()) {
                var = enm.nextElement();
                type = var.getType();
                if(type == EdgeSet.class)
                    fb.addSet((EdgeSet)var.getValue().objectValue());
                else if(type == TupleSet.class)
                    fb.addSet((TupleSet)var.getValue().objectValue());
            }
        } else {
            NodeSet set;
            Node[] nodes;
            
            set = (NodeSet)vals[1].objectValue();
            nodes = set.getAllNodes();
            
            for(int i = 0; i < nodes.length; i++) {
                try {
                    var = env.lookup(nodes[i].get());
                    type = var.getType();
                    if(type == EdgeSet.class)
                        fb.addSet((EdgeSet)var.getValue().objectValue());
                    else if(type == TupleSet.class)
                        fb.addSet((TupleSet)var.getValue().objectValue());
                } catch(LookupException e) {
                    continue;
                }
            }
        }
        
        try {
            writer.write(fileName, fb);
        } catch(IOException e) {
            throw new InvocationException(e.getMessage());
        }
        
        return Value.VOID;
    }

}
