package ca.uwaterloo.cs.jgrok.lib.math;

import java.util.Arrays;
import java.util.HashMap;

import ca.uwaterloo.cs.jgrok.env.Env;
import ca.uwaterloo.cs.jgrok.fb.*;
import ca.uwaterloo.cs.jgrok.interp.*;
import ca.uwaterloo.cs.jgrok.lib.Function;
import ca.uwaterloo.cs.jgrok.lib.InvocationException;

/**
 * <pre>
 * Functions:
 *
 *     set max(set)
 *     set max(set, int)
 *
 *     int maxi(set)
 *     float maxf(set)
 *
 * Examples:
 *     A { age = 10 }
 *     B { age = 10 }
 *     C { age = 30 }
 *     D { age = 32 }
 *     E { age = 32 }
 *
 *     >> max(rng @age)
 *     32
 *     >> max(rng @age, 2)
 *     32
 *     30
 *     >> maxi(rng @age) / 4
 *     8
 *     >> maxf(rng @age) / 4
 *     8.0
 *     >> @age . max(rng @age}
 *     D
 *     E
 *
 * </pre>
 */
public class Max extends Function {
    private static final int METHOD_MAX = 0;
    private static final int METHOD_MAX_INT = 1;
    private static final int MATHOD_MAX_FLOAT = 2;
    private int whichMethod = METHOD_MAX;
    
    public Max(String functionName) {
        this.name = functionName;
        if(this.name.equals("max")) {
            whichMethod = METHOD_MAX;
        } else if(this.name.equals("maxi")) {
            whichMethod = METHOD_MAX_INT;
        } else if(this.name.equals("maxf")) {
            whichMethod = MATHOD_MAX_FLOAT;
        }
        
        type = NodeSet.class;
        paramTypes = new Class[1];
        paramTypes[0] = NodeSet.class;
    }
    
    public Value invoke(Env env, Value[] vals)
        throws InvocationException {
        if(vals.length == 2) {
            return multiset(vals);
        }
        
        float maxFloat = 0;
        NodeSet result = null;
        NodeSet set = (NodeSet)vals[0].objectValue();
            
        Node node;
        Node[] data;
        data = set.getAllNodes();
        if(data.length > 0) {
            int i = 0;
            float val = -1;
            for(; i < data.length; i++) {
                node = data[i];
                try {
                    val = Float.parseFloat(node.get());
                    maxFloat = val;
                    break;
                } catch(NumberFormatException e) {}
            }
            
            if(i < data.length) {
                String s;
                String s_max = data[i].get();
                
                for(; i < data.length; i++) {
                    node = data[i];
                    try {
                        s = node.get();
                        val = Float.parseFloat(s);
                        if(val > maxFloat){
                            s_max = s;
                            maxFloat = val;
                        }
                    } catch(NumberFormatException e) {}
                }
                result = NodeSet.singleton(s_max);
            }
        }
        
        switch(whichMethod) {
        case METHOD_MAX:
            if(result == null) result = new NodeSet();
            return new Value(result);
        case METHOD_MAX_INT:
            if(result == null) break;
            return new Value((int)maxFloat);
        case MATHOD_MAX_FLOAT:
            if(result == null) break;
            return new Value(maxFloat);
        }
        
        return Value.VOID;
    }
    
    private Value multiset(Value[] vals) {
        int count;
        NodeSet set;
        NodeSet result;
        
        set = (NodeSet)vals[0].objectValue();
        count = vals[1].intValue();
        result = new NodeSet();
        
        if(count > 0) {
            Node[] data;
            HashMap<Float,String> map;
            
            data = set.getAllNodes();            
            map = new HashMap<Float,String>(101, 0.75f); 
            
            Float f;
            String s;
            for(int i = 0; i < data.length; i++) {
                try {
                    s = data[i].get();
                    f = new Float(s);
                    map.put(f, s);
                } catch(NumberFormatException e) {}
            }
            
            Object[] keys;
            keys = map.keySet().toArray();
            Arrays.sort(keys);
            
            for(int i = keys.length-1; i >= 0 && count > 0; i--, count--) {
                result.add((String)map.get(keys[i]));
            }
        }
        
        return new Value(result);
    }
}
